#
# Copyright (c) 2018 Joe Todd
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys

__version__ = '0.0.4'


try:
    import _soundiox as soundio
except ImportError:
    print('Please install libsoundio, then reinstall pysoundio')
    sys.exit(-1)

from _soundiox import (
    SoundIoBackendNone, SoundIoBackendJack, SoundIoBackendPulseAudio,
    SoundIoBackendAlsa, SoundIoBackendCoreAudio, SoundIoBackendWasapi,
    SoundIoBackendDummy,
    SoundIoFormatS8, SoundIoFormatU8, SoundIoFormatS16LE,
    SoundIoFormatS16BE, SoundIoFormatU16LE, SoundIoFormatU16BE,
    SoundIoFormatS24LE, SoundIoFormatS24BE, SoundIoFormatU24LE,
    SoundIoFormatU24BE, SoundIoFormatS32LE, SoundIoFormatS32BE,
    SoundIoFormatU32LE, SoundIoFormatU32BE,
    SoundIoFormatFloat32LE, SoundIoFormatFloat32BE, SoundIoFormatFloat64LE,
    SoundIoFormatFloat64BE, SoundIoFormatInvalid
)
from .structures import (
    SoundIo,
    SoundIoChannelArea,
    SoundIoChannelLayout,
    SoundIoDevice,
    SoundIoInStream,
    SoundIoOutStream,
    SoundIoRingBuffer,
)
from .pysoundio import PySoundIo, PySoundIoError



"""
play.py

Stream a wav file to the default output device.
Supports specifying backend, device and block size.

Requires pysoundfile
    pip install pysoundfile

http://pysoundfile.readthedocs.io/
"""
import argparse
import struct
import time

import soundfile as sf
from pysoundio import (
    PySoundIo,
    SoundIoFormatFloat32LE,
    SoundIoOutStream,
)


class Player(object):

    def __init__(self, infile, backend=None, output_device=None, block_size=None):

        data, rate = sf.read(
            infile,
            dtype='float32',
            always_2d=True
        )
        self.data = [d[0] for d in data]
        self.block_size = block_size

        self.pysoundio = PySoundIo(backend=None)
        self.pysoundio.start_output_stream(
            device_id=output_device,
            channels=1,
            sample_rate=rate,
            block_size=self.block_size,
            dtype=SoundIoFormatFloat32LE,
            write_callback=self.callback
        )
        self.cb = 0
        self.total_blocks = len(self.data)
        self.timer = self.total_blocks / float(rate)

    def close(self):
        self.pysoundio.close()

    def callback(self, data, length):
        dlen = (self.block_size if
                self.cb + self.block_size <= self.total_blocks else
                self.total_blocks - self.cb)
        data[:] = struct.pack('%sf' % dlen, *self.data[self.cb:self.cb + dlen])
        self.cb += dlen


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='PySoundIo audio player example',
        epilog='Play a wav file over the default output device'
    )
    parser.add_argument('infile', help='WAV output file name')
    parser.add_argument('--backend', help='Backend to use (optional)')
    parser.add_argument('--blocksize', type=int, default=4096, help='Block size (optional)')
    parser.add_argument('--device', help='Output device id (optional)')
    args = parser.parse_args()

    player = Player(args.infile, args.backend, args.device, args.blocksize)
    print('Playing...')
    print('CTRL-C to exit')

    try:
        time.sleep(player.timer)
    except KeyboardInterrupt:
        print('Exiting...')

    player.close()



"""
constants.py

Constants and enumerations.
"""

from _soundiox import (
    SoundIoBackendNone, SoundIoBackendJack, SoundIoBackendPulseAudio,
    SoundIoBackendAlsa, SoundIoBackendCoreAudio, SoundIoBackendWasapi,
    SoundIoBackendDummy,
    SoundIoFormatS8, SoundIoFormatU8, SoundIoFormatS16LE,
    SoundIoFormatS16BE, SoundIoFormatU16LE, SoundIoFormatU16BE,
    SoundIoFormatS24LE, SoundIoFormatS24BE, SoundIoFormatU24LE,
    SoundIoFormatU24BE, SoundIoFormatS32LE, SoundIoFormatS32BE,
    SoundIoFormatU32LE, SoundIoFormatU32BE,
    SoundIoFormatFloat32LE, SoundIoFormatFloat32BE, SoundIoFormatFloat64LE,
    SoundIoFormatFloat64BE, SoundIoFormatInvalid
)

DEFAULT_RING_BUFFER_DURATION = 30  # secs

SoundIoBackend = {
    SoundIoBackendNone: 'SoundIoBackendNone',
    SoundIoBackendJack: 'SoundIoBackendJack',
    SoundIoBackendPulseAudio: 'SoundIoBackendPulseAudio',
    SoundIoBackendAlsa: 'SoundIoBackendAlsa',
    SoundIoBackendCoreAudio: 'SoundIoBackendCoreAudio',
    SoundIoBackendWasapi: 'SoundIoBackendWasapi',
    SoundIoBackendDummy: 'SoundIoBackendDummy',
}

SoundIoFormat = {
    SoundIoFormatS8: 'SoundIoFormatS8',
    SoundIoFormatU8: 'SoundIoFormatU8',
    SoundIoFormatS16LE: 'SoundIoFormatS16LE',
    SoundIoFormatS16BE: 'SoundIoFormatS16BE',
    SoundIoFormatU16LE: 'SoundIoFormatU16LE',
    SoundIoFormatU16BE: 'SoundIoFormatU16BE',
    SoundIoFormatS24LE: 'SoundIoFormatS24LE',
    SoundIoFormatS24BE: 'SoundIoFormatS24BE',
    SoundIoFormatU24LE: 'SoundIoFormatU24LE',
    SoundIoFormatU24BE: 'SoundIoFormatU24BE',
    SoundIoFormatS32LE: 'SoundIoFormatS32LE',
    SoundIoFormatS32BE: 'SoundIoFormatS32BE',
    SoundIoFormatU32LE: 'SoundIoFormatU32LE',
    SoundIoFormatU32BE: 'SoundIoFormatU32BE',
    SoundIoFormatFloat32LE: 'SoundIoFormatFloat32LE',
    SoundIoFormatFloat32BE: 'SoundIoFormatFloat32BE',
    SoundIoFormatFloat64LE: 'SoundIoFormatFloat64LE',
    SoundIoFormatFloat64BE: 'SoundIoFormatFloat64BE',
    SoundIoFormatInvalid: 'SoundIoFormatInvalid'
}

PRIORITISED_FORMATS = [
    SoundIoFormatFloat32LE,
    SoundIoFormatFloat32BE,
    SoundIoFormatS32LE,
    SoundIoFormatS32BE,
    SoundIoFormatS24LE,
    SoundIoFormatS24BE,
    SoundIoFormatS16LE,
    SoundIoFormatS16BE,
    SoundIoFormatFloat64LE,
    SoundIoFormatFloat64BE,
    SoundIoFormatU32LE,
    SoundIoFormatU32BE,
    SoundIoFormatU24LE,
    SoundIoFormatU24BE,
    SoundIoFormatU16LE,
    SoundIoFormatU16BE,
    SoundIoFormatS8,
    SoundIoFormatU8,
    SoundIoFormatInvalid,
]

PRIORITISED_SAMPLE_RATES = [
    48000,
    44100,
    96000,
    24000,
    0,
]

ARRAY_FORMATS = {
    SoundIoFormatS8: 'b',
    SoundIoFormatU8: 'B',
    SoundIoFormatS16LE: 'h',
    SoundIoFormatS16BE: 'h',
    SoundIoFormatU16LE: 'H',
    SoundIoFormatU16BE: 'H',
    SoundIoFormatS24LE: None,
    SoundIoFormatS24BE: None,
    SoundIoFormatU24LE: None,
    SoundIoFormatU24BE: None,
    SoundIoFormatS32LE: 'l',
    SoundIoFormatS32BE: 'l',
    SoundIoFormatU32LE: 'L',
    SoundIoFormatU32BE: 'L',
    SoundIoFormatFloat32LE: 'f',
    SoundIoFormatFloat32BE: 'f',
    SoundIoFormatFloat64LE: 'd',
    SoundIoFormatFloat64BE: 'd'
}

SOUNDFILE_FORMATS = {
    SoundIoFormatS8: None,
    SoundIoFormatU8: None,
    SoundIoFormatS16LE: 'int16',
    SoundIoFormatS16BE: 'int16',
    SoundIoFormatU16LE: None,
    SoundIoFormatU16BE: None,
    SoundIoFormatS24LE: None,
    SoundIoFormatS24BE: None,
    SoundIoFormatU24LE: None,
    SoundIoFormatU24BE: None,
    SoundIoFormatS32LE: 'int32',
    SoundIoFormatS32BE: 'int32',
    SoundIoFormatU32LE: None,
    SoundIoFormatU32BE: None,
    SoundIoFormatFloat32LE: 'float32',
    SoundIoFormatFloat32BE: 'float32',
    SoundIoFormatFloat64LE: 'float64',
    SoundIoFormatFloat64BE: 'float64'
}


"""
structures.py

Implements data structures and callbacks from libsoundio
"""
import ctypes as _ctypes
from _soundiox import SOUNDIO_MAX_CHANNELS


SoundIoReadCallback = _ctypes.CFUNCTYPE(
    None,
    _ctypes.c_void_p, _ctypes.c_int, _ctypes.c_int,
)
SoundIoWriteCallback = _ctypes.CFUNCTYPE(
    None,
    _ctypes.c_void_p, _ctypes.c_int, _ctypes.c_int,
)
SoundIoOverflowCallback = _ctypes.CFUNCTYPE(
    None,
    _ctypes.c_void_p
)
SoundIoUnderflowCallback = _ctypes.CFUNCTYPE(
    None,
    _ctypes.c_void_p
)
SoundIoErrorCallback = _ctypes.CFUNCTYPE(
    None,
    _ctypes.c_void_p, _ctypes.c_int
)

class SoundIo(_ctypes.Structure):
    """
    Forward declaration of SoundIo structure
    """
    pass


_SoundIoOnDevicesChange = _ctypes.CFUNCTYPE(
    None,
    _ctypes.POINTER(SoundIo)
)
_SoundIoOnBackendDisconnect = _ctypes.CFUNCTYPE(
    None,
    _ctypes.POINTER(SoundIo), _ctypes.c_int
)
_SoundIoOnEventsSignal = _ctypes.CFUNCTYPE(
    None,
    _ctypes.POINTER(SoundIo)
)
_SoundIoEmitRtprioWarning = _ctypes.CFUNCTYPE(None)
_SoundIoJackInfoCallback = _ctypes.CFUNCTYPE(
    None,
    _ctypes.c_char_p
)
_SoundIoJackErrorCallback = _ctypes.CFUNCTYPE(
    None,
    _ctypes.c_char_p
)


SoundIo._fields_ = [
    ('userdata', _ctypes.c_void_p),
    ('on_devices_change', _SoundIoOnDevicesChange),
    ('on_backend_disconnect', _SoundIoOnBackendDisconnect),
    ('on_events_signal', _SoundIoOnEventsSignal),
    ('current_backend', _ctypes.c_uint),
    ('app_name', _ctypes.c_char_p),
    ('emit_rtprio_warning', _SoundIoEmitRtprioWarning),
    ('jack_info_callback', _SoundIoJackInfoCallback),
    ('jack_error_callback', _SoundIoJackErrorCallback)
]


class SoundIoChannelArea(_ctypes.Structure):
    _fields_ = [
        ('ptr', _ctypes.POINTER(_ctypes.c_char)),
        ('step', _ctypes.c_int)
    ]


class SoundIoSampleRateRange(_ctypes.Structure):
    _fields_ = [
        ('min', _ctypes.c_int),
        ('max', _ctypes.c_int),
    ]


class SoundIoChannelLayout(_ctypes.Structure):
    _fields_ = [
        ('name', _ctypes.c_char_p),
        ('channel_count', _ctypes.c_int),
        ('channels', _ctypes.c_uint * SOUNDIO_MAX_CHANNELS)
    ]


class SoundIoDevice(_ctypes.Structure):
    _fields_ = [
        ('soundio', _ctypes.POINTER(SoundIo)),
        ('id', _ctypes.c_char_p),
        ('name', _ctypes.c_char_p),
        ('aim', _ctypes.c_uint),
        ('layouts', _ctypes.POINTER(SoundIoChannelLayout)),
        ('layout_count', _ctypes.c_int),
        ('current_layout', SoundIoChannelLayout),
        ('formats', _ctypes.POINTER(_ctypes.c_uint)),
        ('format_count', _ctypes.c_int),
        ('current_format', _ctypes.c_uint),
        ('sample_rates', _ctypes.POINTER(SoundIoSampleRateRange)),
        ('sample_rate_count', _ctypes.c_int),
        ('sample_rate_current', _ctypes.c_int),
        ('software_latency_min', _ctypes.c_double),
        ('software_latency_max', _ctypes.c_double),
        ('software_latency_current', _ctypes.c_double),
        ('is_raw', _ctypes.c_bool),
        ('ref_count', _ctypes.c_int),
        ('probe_error', _ctypes.c_int)
    ]


class SoundIoInStream(_ctypes.Structure):
    _fields_ = [
        ('device', _ctypes.POINTER(SoundIoDevice)),
        ('format', _ctypes.c_uint),
        ('sample_rate', _ctypes.c_int),
        ('layout', SoundIoChannelLayout),
        ('software_latency', _ctypes.c_double),
        ('userdata', _ctypes.c_void_p),
        ('read_callback', SoundIoReadCallback),
        ('overflow_callback', SoundIoOverflowCallback),
        ('error_callback', SoundIoErrorCallback),
        ('name', _ctypes.c_char_p),
        ('non_terminal_hint', _ctypes.c_bool),
        ('bytes_per_frame', _ctypes.c_int),
        ('bytes_per_sample', _ctypes.c_int),
        ('layout_error', _ctypes.c_int)
    ]


class SoundIoOutStream(_ctypes.Structure):
    _fields_ = [
        ('device', _ctypes.POINTER(SoundIoDevice)),
        ('format', _ctypes.c_uint),
        ('sample_rate', _ctypes.c_int),
        ('layout', SoundIoChannelLayout),
        ('software_latency', _ctypes.c_double),
        ('userdata', _ctypes.c_void_p),
        ('write_callback', SoundIoWriteCallback),
        ('underflow_callback', SoundIoUnderflowCallback),
        ('error_callback', SoundIoErrorCallback),
        ('name', _ctypes.c_char_p),
        ('non_terminal_hint', _ctypes.c_bool),
        ('bytes_per_frame', _ctypes.c_int),
        ('bytes_per_sample', _ctypes.c_int),
        ('layout_error', _ctypes.c_int)
    ]


class _SoundIoOsMirroredMemory(_ctypes.Structure):
    _fields_ = [
        ('capacity', _ctypes.c_size_t),
        ('address', _ctypes.c_char_p),
        ('priv', _ctypes.c_void_p)
    ]


class _SoundIoAtomicULong(_ctypes.Structure):
    _fields_ = [
        ('x', _ctypes.c_ulong)
    ]


class SoundIoRingBuffer(_ctypes.Structure):
    _fields_ = [
        ('mem', _SoundIoOsMirroredMemory),
        ('write_offset', _SoundIoAtomicULong),
        ('read_offset', _SoundIoAtomicULong),
        ('capacity', _ctypes.c_int)
    ]



"""
record.py

Stream the default input device and save to wav file.
Supports specifying backend, device, sample rate, block size.

Requires pysoundfile
    pip install pysoundfile

http://pysoundfile.readthedocs.io/
"""
import argparse
import time

import soundfile as sf
from pysoundio import (
    PySoundIo,
    SoundIoFormatFloat32LE,
)


class Record(object):

    def __init__(self, outfile, backend=None, input_device=None,
                 sample_rate=None, block_size=None, channels=None):
        self.wav_file = sf.SoundFile(
            outfile, mode='w', channels=channels,
            samplerate=sample_rate
        )
        self.pysoundio = PySoundIo(backend=None)
        self.pysoundio.start_input_stream(
            device_id=input_device,
            channels=channels,
            sample_rate=sample_rate,
            block_size=block_size,
            dtype=SoundIoFormatFloat32LE,
            read_callback=self.callback
        )

    def close(self):
        self.pysoundio.close()
        self.wav_file.close()

    def callback(self, data, length):
        self.wav_file.buffer_write(data, dtype='float32')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='PySoundIo audio record example',
        epilog='Stream the input device and save to wav file'
    )
    parser.add_argument('outfile', help='WAV output file name')
    parser.add_argument('--backend', help='Backend to use (optional)')
    parser.add_argument('--blocksize', help='Block size (optional)')
    parser.add_argument('--rate', default=44100, help='Sample rate (optional)')
    parser.add_argument('--channels', type=int, default=2, help='Mono or stereo (optional)')
    parser.add_argument('--device', help='Input device id (optional)')
    args = parser.parse_args()

    record = Record(args.outfile, args.backend, args.device, args.rate, args.blocksize, args.channels)
    print('Recording...')
    print('CTRL-C to exit')

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print('Exiting...')

    record.close()



"""
test_pysoundio.py

PySoundIo Test Suite
"""
import unittest
import pysoundio
import _soundiox


class TestPySoundIo(unittest.TestCase):

    def setUp(self):
        self.sio = pysoundio.PySoundIo(
            backend=pysoundio.SoundIoBackendDummy)
        self.sio.input['channels'] = 2
        self.sio.input['sample_rate'] = 44100
        self.sio.input['format'] = pysoundio.SoundIoFormatFloat32LE
        self.sio.input['block_size'] = None

        self.sio.output['channels'] = 2
        self.sio.output['sample_rate'] = 44100
        self.sio.output['format'] = pysoundio.SoundIoFormatFloat32LE
        self.sio.output['block_size'] = None

        self.sio.testing = True
        self.callback_called = False

    def tearDown(self):
        self.sio.close()

    # - Device API

    def test_backend_count(self):
        self.assertIsInstance(self.sio.backend_count, int)

    def test_get_default_input_device(self):
        self.sio.input['device'] = self.sio.get_default_input_device()
        self.assertIsNotNone(self.sio.input['device'])

    def test_get_input_device(self):
        self.sio.input['device'] = self.sio.get_input_device(0)
        self.assertIsNotNone(self.sio.input['device'])

    def test_invalid_input_device(self):
        with self.assertRaises(pysoundio.PySoundIoError):
            self.sio.get_input_device(100)

    def test_get_default_output_device(self):
        self.sio.output['device'] = self.sio.get_default_output_device()
        self.assertIsNotNone(self.sio.output['device'])

    def test_get_output_device(self):
        self.sio.output['device'] = self.sio.get_output_device(0)
        self.assertIsNotNone(self.sio.output['device'])

    def test_invalid_output_device(self):
        with self.assertRaises(pysoundio.PySoundIoError):
            self.sio.get_output_device(100)

    def test_get_layouts(self):
        self.sio.input['device'] = self.sio.get_default_input_device()
        layouts = self.sio.get_layouts(self.sio.input['device'])
        self.assertIsInstance(layouts, dict)
        self.assertIn('current', layouts)
        self.assertIn('available', layouts)
        self.assertTrue(len(layouts['available']) > 0)
        self.assertIn('name', layouts['available'][0])
        self.assertIn('channel_count', layouts['available'][0])

    def test_get_sample_rates(self):
        self.sio.input['device'] = self.sio.get_default_input_device()
        rates = self.sio.get_sample_rates(self.sio.input['device'])
        self.assertIsInstance(rates, dict)
        self.assertIn('current', rates)
        self.assertIn('available', rates)
        self.assertTrue(len(rates['available']) > 0)
        self.assertIn('min', rates['available'][0])
        self.assertIn('max', rates['available'][0])

    def test_get_formats(self):
        self.sio.input['device'] = self.sio.get_default_input_device()
        formats = self.sio.get_formats(self.sio.input['device'])
        self.assertIsInstance(formats, dict)
        self.assertIn('current', formats)
        self.assertIn('available', formats)
        self.assertTrue(len(formats['available']) > 0)

    def test_list_devices(self):
        input_devices, output_devices = self.sio.list_devices()
        self.assertIsInstance(input_devices, list)
        self.assertIsInstance(output_devices, list)
        self.assertTrue(len(input_devices) > 0)
        self.assertTrue(len(output_devices) > 0)
        self.assertIsInstance(input_devices[0], dict)
        self.assertIsInstance(output_devices[0], dict)
        for device in [input_devices[0], output_devices[0]]:
            self.assertIn('id', device)
            self.assertIn('name', device)
            self.assertIn('is_raw', device)
            self.assertIn('sample_rates', device)
            self.assertIn('formats', device)
            self.assertIn('layouts', device)
            self.assertIn('software_latency_min', device)
            self.assertIn('software_latency_max', device)
            self.assertIn('software_latency_current', device)

    def test_supports_sample_rate(self):
        self.sio.input['device'] = self.sio.get_input_device(0)
        self.assertTrue(self.sio.supports_sample_rate(self.sio.input['device'], 44100))

    def test_get_default_sample_rate(self):
        self.sio.input['device'] = self.sio.get_input_device(0)
        self.assertIsInstance(self.sio.get_default_sample_rate(self.sio.input['device']), int)

    def test_get_default_sample_rate_max(self):
        sample_rates = pysoundio.pysoundio.PRIORITISED_SAMPLE_RATES
        pysoundio.pysoundio.PRIORITISED_SAMPLE_RATES = [0]
        self.sio.input['device'] = self.sio.get_input_device(0)
        self.assertIsInstance(self.sio.get_default_sample_rate(self.sio.input['device']), int)
        pysoundio.pysoundio.PRIORITISED_SAMPLE_RATES = sample_rates

    def test_supports_format(self):
        self.sio.input['device'] = self.sio.get_input_device(0)
        self.assertTrue(self.sio.supports_format(self.sio.input['device'], pysoundio.SoundIoFormatFloat32LE))

    def test_get_default_format(self):
        self.sio.input['device'] = self.sio.get_input_device(0)
        self.assertIsInstance(self.sio.get_default_format(self.sio.input['device']), int)

    def test_get_default_format_invalid(self):
        formats = pysoundio.pysoundio.PRIORITISED_FORMATS
        pysoundio.pysoundio.PRIORITISED_FORMATS = []
        self.sio.input['device'] = self.sio.get_input_device(0)
        with self.assertRaises(pysoundio.PySoundIoError):
            self.assertIsInstance(self.sio.get_default_format(self.sio.input['device']), int)
        pysoundio.pysoundio.PRIORITISED_FORMATS = formats

    def test_get_default_layout(self):
        self.assertIsNotNone(self.sio._get_default_layout(2))

    def test_bytes_per_frame(self):
        self.assertEqual(self.sio.get_bytes_per_frame(
            pysoundio.SoundIoFormatFloat32LE, 2), 8)

    def test_bytes_per_sample(self):
        self.assertEqual(self.sio.get_bytes_per_sample(
            pysoundio.SoundIoFormatFloat32LE), 4)

    def test_bytes_per_second(self):
        self.assertEqual(self.sio.get_bytes_per_second(
            pysoundio.SoundIoFormatFloat32LE, 1, 44100), 176400)

    # - Input Stream API

    def fill_input_buffer(self):
        data = bytearray(b'\x00' * 4096 * 8)
        _soundiox.ring_buffer_write_ptr(self.sio.input['buffer'], data, len(data))
        _soundiox.ring_buffer_advance_write_ptr(self.sio.input['buffer'], len(data))

    def test__create_input_ring_buffer(self):
        capacity = 44100 * 8
        self.sio.input['buffer'] = self.sio._create_input_ring_buffer(capacity)
        self.assertIsNotNone(self.sio.input['buffer'])

    def test_get_input_latency(self):
        self.sio.start_input_stream(
            sample_rate=44100,
            dtype=pysoundio.SoundIoFormatFloat32LE,
            channels=2)
        self.fill_input_buffer()
        self.assertIsInstance(self.sio.get_input_latency(0.2), int)

    def overflow_callback(self):
        self.callback_called = True

    def test_overflow_callback(self):
        self.sio.input['overflow_callback'] = self.overflow_callback
        self.sio._overflow_callback()
        self.assertTrue(self.callback_called)

    def test_start_input_stream(self):
        self.sio.start_input_stream(
            sample_rate=44100,
            dtype=pysoundio.SoundIoFormatFloat32LE,
            channels=2)
        self.fill_input_buffer()
        self.assertIsNotNone(self.sio.input['stream'])

    def test_pause_input_stream(self):
        self.sio.start_input_stream(
            sample_rate=44100,
            dtype=pysoundio.SoundIoFormatFloat32LE,
            channels=2)
        self.fill_input_buffer()
        self.sio.pause_input_stream(True)

    def test_start_input_stream_device(self):
        self.sio.start_input_stream(
            device_id=0,
            sample_rate=44100,
            dtype=pysoundio.SoundIoFormatFloat32LE,
            channels=2)
        self.fill_input_buffer()
        self.assertIsNotNone(self.sio.input['stream'])

    def test_start_input_invalid_rate(self):
        with self.assertRaises(pysoundio.PySoundIoError):
            self.sio.start_input_stream(
                sample_rate=10000000,
                dtype=pysoundio.SoundIoFormatFloat32LE,
                channels=2)

    def test_start_input_default_rate(self):
        self.sio.start_input_stream(
            dtype=pysoundio.SoundIoFormatFloat32LE,
            channels=2)
        self.fill_input_buffer()
        self.assertIsNotNone(self.sio.input['stream'])
        self.assertIsInstance(self.sio.input['sample_rate'], int)

    def test_start_input_invalid_format(self):
        with self.assertRaises(pysoundio.PySoundIoError):
            self.sio.start_input_stream(
                sample_rate=44100,
                dtype=50,
                channels=2)

    def test_start_input_default_format(self):
        self.sio.start_input_stream(
            sample_rate=44100,
            channels=2)
        self.fill_input_buffer()
        self.assertIsNotNone(self.sio.input['stream'])
        self.assertIsInstance(self.sio.input['format'], int)

    # -- Output Stream API

    def test__create_output_ring_buffer(self):
        capacity = 44100 * 8
        self.assertIsNotNone(self.sio._create_output_ring_buffer(capacity))

    def test__open_output_stream(self):
        self.sio.output['device'] = self.sio.get_default_output_device()
        self.sio.output['stream'] = self.sio._create_output_stream()
        self.sio._open_output_stream()
        self.assertIsNotNone(self.sio.output['block_size'])

    def test__open_output_stream_blocksize(self):
        self.sio.block_size = 4096
        self.sio.output['device'] = self.sio.get_default_output_device()
        self.sio.output['stream'] = self.sio._create_output_stream()
        self.sio._open_output_stream()
        self.assertIsNotNone(self.sio.output['block_size'])

    def test_clear_output_buffer(self):
        capacity = 44100 * 8
        self.sio.output['buffer'] = self.sio._create_output_ring_buffer(capacity)
        self.assertIsNotNone(self.sio.output['buffer'])
        self.sio._clear_output_buffer()

    def underflow_callback(self):
        self.callback_called = True

    def test_underflow_callback(self):
        self.sio.output['underflow_callback'] = self.underflow_callback
        self.sio._underflow_callback()
        self.assertTrue(self.callback_called)

    def test_get_output_latency(self):
        self.sio.start_output_stream(
            sample_rate=44100,
            dtype=pysoundio.SoundIoFormatFloat32LE,
            channels=2)
        self.assertIsInstance(self.sio.get_output_latency(0.2), int)

    def test_pause_output_stream(self):
        self.sio.start_output_stream(
            sample_rate=44100,
            dtype=pysoundio.SoundIoFormatFloat32LE,
            channels=2)
        self.sio.pause_output_stream(True)

    def test_start_output_stream(self):
        self.sio.start_output_stream(
            sample_rate=44100,
            dtype=pysoundio.SoundIoFormatFloat32LE,
            channels=2)
        self.assertIsNotNone(self.sio.output['stream'])

    def test_start_output_stream_device(self):
        self.sio.start_output_stream(
            device_id=0,
            sample_rate=44100,
            dtype=pysoundio.SoundIoFormatFloat32LE,
            channels=2)
        self.assertIsNotNone(self.sio.output['stream'])

    def test_start_output_invalid_rate(self):
        with self.assertRaises(pysoundio.PySoundIoError):
            self.sio.start_output_stream(
                sample_rate=10000000,
                dtype=pysoundio.SoundIoFormatFloat32LE,
                channels=2)

    def test_start_output_default_rate(self):
        self.sio.start_output_stream(
            dtype=pysoundio.SoundIoFormatFloat32LE,
            channels=2)
        self.assertIsNotNone(self.sio.output['stream'])
        self.assertIsInstance(self.sio.output['sample_rate'], int)

    def test_start_output_invalid_format(self):
        with self.assertRaises(pysoundio.PySoundIoError):
            self.sio.start_output_stream(
                sample_rate=44100,
                dtype=50,
                channels=2)

    def test_start_output_default_format(self):
        self.sio.start_output_stream(
            sample_rate=44100,
            channels=2)
        self.assertIsNotNone(self.sio.output['stream'])
        self.assertIsInstance(self.sio.output['format'], int)


class TestInputProcessing(unittest.TestCase):

    def setUp(self):
        self.sio = pysoundio.PySoundIo(
            backend=pysoundio.SoundIoBackendDummy)
        self.callback_called = False

    def tearDown(self):
        self.sio.close()

    def callback(self, data, length):
        self.callback_called = True

    def test_read_callback(self):
        self.sio.start_input_stream(
            sample_rate=44100,
            dtype=pysoundio.SoundIoFormatFloat32LE,
            channels=2,
            block_size=4096,
            read_callback=self.callback)
        self.assertIsNotNone(self.sio.input['stream'])

        data = bytearray(b'\x00' * 4096 * 8)
        _soundiox.ring_buffer_write_ptr(self.sio.input['buffer'], data, len(data))
        _soundiox.ring_buffer_advance_write_ptr(self.sio.input['buffer'], len(data))

        thread = pysoundio.pysoundio._InputProcessingThread(parent=self.sio)
        thread.run()
        self.assertTrue(self.callback_called)


class TestOutputProcessing(unittest.TestCase):

    def setUp(self):
        self.sio = pysoundio.PySoundIo(
            backend=pysoundio.SoundIoBackendDummy)
        self.sio.testing = True
        self.callback_called = False

    def tearDown(self):
        self.sio.close()

    def callback(self, data, length):
        self.callback_called = True

    def test_write_callback(self):
        self.sio.start_output_stream(
            sample_rate=44100,
            dtype=pysoundio.SoundIoFormatFloat32LE,
            channels=2,
            block_size=4096,
            write_callback=self.callback)
        self.assertIsNotNone(self.sio.output['stream'])
        thread = pysoundio.pysoundio._OutputProcessingThread(parent=self.sio, block_size=4096)
        thread.run()
        self.assertTrue(self.callback_called)



"""
devices.py

List the available input and output devices on the system and their properties.
"""
import argparse

from pysoundio import PySoundIo
from pysoundio.constants import SoundIoFormat


def print_devices(devices):
    for i, device in enumerate(devices):
        msg = '* ' if device['is_default'] else '  '
        msg += str(i) + ' - ' + device['name']
        print(msg)
        print('\tsample rate:')
        print('\t default: {}Hz'.format(device['sample_rates']['current']))
        print('\t available: {}'.format(
            ', '.join([str(d['max']) + 'Hz' for d in device['sample_rates']['available']])))
        print('\tformat: {}'.format(
            ', '.join([str(d) for d in device['formats']['available']])))
        print('\tlayouts: {}'.format(device['layouts']['current']['name']))
        print('\t available: {}'.format(
            ', '.join([str(d['name']) for d in device['layouts']['available']])))
        print('\tsoftware latency:')
        print('\t min: {}s, max: {}s, current: {}s'.format(
            round(device['software_latency_min'], 4),
            round(device['software_latency_max'], 4),
            round(device['software_latency_current'], 4)))
        print("")


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='PySoundIo list devices example',
        epilog='List the available input and output devices'
    )
    parser.add_argument('--backend', help='Backend to use (optional)')
    args = parser.parse_args()

    pysoundio = PySoundIo(backend=args.backend)
    input_devices, output_devices = pysoundio.list_devices()
    print('\n' + '-' * 20 + ' Input Devices ' + '-' * 20 + '\n')
    print_devices(input_devices)

    print('\n' + '-' * 20 + ' Output Devices ' + '-' * 20 + '\n')
    print_devices(output_devices)

    pysoundio.close()



"""
test_soundiox.py

C API Test Suite
"""
import ctypes
import unittest
import pysoundio
import _soundiox as soundio


class TestInitialisation(unittest.TestCase):

    def setUp(self):
        self.s = soundio.create()

    def tearDown(self):
        soundio.destroy()

    def test_connect(self):
        self.assertEqual(
            soundio.connect_backend(pysoundio.SoundIoBackendDummy), 0)

    def test_version(self):
        self.assertIsInstance(soundio.version_string(), str)

    def test_format_string(self):
        self.assertEqual(
            soundio.format_string(pysoundio.SoundIoFormatFloat32LE),
            'float 32-bit LE'
        )

    def test_backend_count(self):
        self.assertIsInstance(soundio.backend_count(), int)


class TestDeviceAPI(unittest.TestCase):

    def setUp(self):
        self.device = None
        self.s = soundio.create()
        soundio.connect_backend(pysoundio.SoundIoBackendDummy)
        soundio.flush()

    def tearDown(self):
        if self.device:
            soundio.device_unref(self.device)
        soundio.destroy()

    def test_get_input_device_count(self):
        self.assertNotEqual(soundio.get_input_device_count(), -1)

    def test_get_output_device_count(self):
        self.assertNotEqual(soundio.get_output_device_count(), -1)

    def test_default_input_device_index(self):
        self.assertNotEqual(soundio.default_input_device_index(), -1)

    def test_default_output_device_index(self):
        self.assertNotEqual(soundio.default_output_device_index(), -1)

    def test_get_input_device(self):
        default_index = soundio.default_input_device_index()
        self.device = soundio.get_input_device(default_index)
        self.assertIsNotNone(self.device)

    def test_get_output_device(self):
        default_index = soundio.default_output_device_index()
        self.device = soundio.get_output_device(default_index)
        self.assertIsNotNone(self.device)

    def test_device_supports_sample_rate(self):
        default_index = soundio.default_output_device_index()
        self.device = soundio.get_output_device(default_index)
        self.assertIsNotNone(soundio.device_supports_sample_rate(self.device, 44100))

    def test_device_supports_format(self):
        default_index = soundio.default_output_device_index()
        self.device = soundio.get_output_device(default_index)
        self.assertIsNotNone(soundio.device_supports_format(
            self.device, pysoundio.SoundIoFormatFloat32LE))

    def test_device_sort_channel_layouts(self):
        default_index = soundio.default_output_device_index()
        self.device = soundio.get_output_device(default_index)
        soundio.device_sort_channel_layouts(self.device)

    def test_channel_layout_get_default(self):
        self.assertIsNotNone(soundio.channel_layout_get_default(2))

    def test_channel_layout_detect_builtin(self):
        layout = soundio.channel_layout_get_default(2)
        self.assertTrue(soundio.channel_layout_detect_builtin(layout))

    def test_channel_layout_equal(self):
        layout = soundio.channel_layout_get_default(2)
        self.assertTrue(soundio.channel_layout_equal(layout, layout))

    def test_channel_layout_find_channel(self):
        layout = soundio.channel_layout_get_default(2)
        self.assertIsInstance(soundio.channel_layout_find_channel(layout, 0), int)

    def test_channel_layout_get_builtin(self):
        self.assertIsNotNone(soundio.channel_layout_get_builtin(0))

    def test_force_device_scan(self):
        soundio.force_device_scan()

    def test_bytes_per_frame(self):
        self.assertEqual(soundio.get_bytes_per_frame(
            pysoundio.SoundIoFormatFloat32LE, 2), 8)

    def test_bytes_per_sample(self):
        self.assertEqual(soundio.get_bytes_per_sample(
            pysoundio.SoundIoFormatFloat32LE), 4)

    def test_bytes_per_second(self):
        self.assertEqual(soundio.get_bytes_per_second(
            pysoundio.SoundIoFormatFloat32LE, 1, 44100), 176400)


class TestInputStreamAPI(unittest.TestCase):

    def setUp(self):
        self.instream = None
        self.s = soundio.create()
        soundio.connect_backend(pysoundio.SoundIoBackendDummy)
        soundio.flush()
        default_index = soundio.default_input_device_index()
        self.device = soundio.get_input_device(default_index)
        self.buffer = soundio.input_ring_buffer_create(44100 * 8)

    def tearDown(self):
        if self.instream:
            soundio.instream_destroy()
        if self.device:
            soundio.device_unref(self.device)
        soundio.destroy()

    def callback(self):
        pass

    def setup_stream(self):
        soundio.set_read_callbacks(self.callback, self.callback)
        self.instream = soundio.instream_create(self.device)
        instream = ctypes.cast(self.instream, ctypes.POINTER(pysoundio.SoundIoInStream))
        instream.contents.format = soundio.SoundIoFormatFloat32LE
        instream.contents.sample_rate = 44100

    def test_instream_create(self):
        self.assertIsNotNone(soundio.instream_create(self.device))

    def test_instream_open(self):
        self.setup_stream()
        self.assertEqual(soundio.instream_open(), 0)

    def test_instream_start(self):
        self.setup_stream()
        soundio.instream_open()
        self.assertEqual(soundio.instream_start(), 0)

    def test_instream_pause(self):
        self.setup_stream()
        soundio.instream_open()
        self.assertEqual(soundio.instream_pause(True), 0)

    def test_instream_get_latency(self):
        self.setup_stream()
        soundio.instream_open()
        self.assertIsInstance(soundio.instream_get_latency(0.42), int)


class TestOutputStreamAPI(unittest.TestCase):

    def setUp(self):
        self.outstream = None
        self.s = soundio.create()
        soundio.connect_backend(pysoundio.SoundIoBackendDummy)
        soundio.flush()
        default_index = soundio.default_output_device_index()
        self.device = soundio.get_output_device(default_index)
        self.buffer = soundio.input_ring_buffer_create(44100 * 8)

    def tearDown(self):
        if self.outstream:
            soundio.outstream_pause(True)
            soundio.outstream_destroy()
        if self.device:
            soundio.device_unref(self.device)
        soundio.disconnect()
        soundio.destroy()

    def setup_stream(self):
        self.outstream = soundio.outstream_create(self.device)
        outstream = ctypes.cast(self.outstream, ctypes.POINTER(pysoundio.SoundIoOutStream))
        outstream.contents.format = soundio.SoundIoFormatFloat32LE
        outstream.contents.sample_rate = 44100
        outstream.contents.software_latency = 0.0

    def test_outstream_create(self):
        self.assertIsNotNone(soundio.outstream_create(self.device))

    def test_outstream_open(self):
        self.setup_stream()
        self.assertEqual(soundio.outstream_open(), 0)

    def test_outstream_pause(self):
        self.setup_stream()
        soundio.outstream_open()
        self.assertEqual(soundio.outstream_pause(True), 0)

    def test_outstream_get_latency(self):
        self.setup_stream()
        soundio.outstream_open()
        self.assertIsInstance(soundio.outstream_get_latency(0.42), int)


class TestRingBufferAPI(unittest.TestCase):

    def setUp(self):
        self.s = soundio.create()
        soundio.connect_backend(pysoundio.SoundIoBackendDummy)
        soundio.flush()
        self.buffer = soundio.input_ring_buffer_create(44100)

        # Fill with some data for libsoundio < v1.1.0
        data = bytearray(b'' * 42)
        soundio.ring_buffer_write_ptr(self.buffer, data, len(data))
        soundio.ring_buffer_advance_write_ptr(self.buffer, 42)

    def tearDown(self):
        if self.buffer:
            soundio.ring_buffer_destroy(self.buffer)
        soundio.destroy()

    def test_ring_buffer_fill_count(self):
        self.assertEqual(soundio.ring_buffer_fill_count(self.buffer), 42)

    def test_ring_buffer_read_ptr(self):
        ptr = soundio.ring_buffer_read_ptr(self.buffer)
        self.assertIsInstance(ptr, bytes)
        self.assertNotEqual(ptr, 0)

    def test_ring_buffer_advance_read_ptr(self):
        count = soundio.ring_buffer_free_count(self.buffer)
        soundio.ring_buffer_advance_read_ptr(self.buffer, 16)
        self.assertEqual(soundio.ring_buffer_free_count(self.buffer), count + 16)

    def test_ring_buffer_write_ptr(self):
        data = bytearray('\x01\x02\x03\x04'.encode())
        soundio.ring_buffer_write_ptr(self.buffer, data, len(data))

    def test_ring_buffer_free_count(self):
        self.assertNotEqual(soundio.ring_buffer_free_count(self.buffer), 0)

    def test_ring_buffer_advance_write_ptr(self):
        self.assertEqual(soundio.ring_buffer_fill_count(self.buffer), 42)
        soundio.ring_buffer_advance_write_ptr(self.buffer, 16)
        self.assertEqual(soundio.ring_buffer_fill_count(self.buffer), 58)

    def test_ring_buffer_clear(self):
        soundio.ring_buffer_clear(self.buffer)

    def test_ring_buffer_capacity(self):
        self.assertIsInstance(soundio.ring_buffer_capacity(self.buffer), int)


if __name__ == '__main__':
    unittest.main()



"""
sine.py

Play a sine wave over the default output device.
Supports specifying backend, device, sample rate, block size.

"""
import array as ar
import argparse
import math
import struct
import time

import _soundiox as soundio
from pysoundio import (
    PySoundIo,
    SoundIoFormatFloat32LE,
    SoundIoOutStream,
)


class Player(object):

    def __init__(self, freq=None, backend=None, output_device=None,
                 sample_rate=None, block_size=None):
        self.pysoundio = PySoundIo(backend=None)

        self.freq = float(freq)
        self.seconds_offset = 0.0
        self.radians_per_second = self.freq * 2.0 * math.pi
        self.seconds_per_frame = 1.0 / sample_rate

        self.pysoundio.start_output_stream(
            device_id=output_device,
            channels=1,
            sample_rate=sample_rate,
            block_size=block_size,
            dtype=SoundIoFormatFloat32LE,
            write_callback=self.callback
        )

    def close(self):
        self.pysoundio.close()

    def callback(self, data, length):
        indata = ar.array('f', [0] * length)
        for i in range(0, length):
            indata[i] = math.sin(
                (self.seconds_offset + i * self.seconds_per_frame) * self.radians_per_second)
        data[:] = indata.tostring()
        self.seconds_offset += self.seconds_per_frame * length


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='PySoundIo sine wave output example',
        epilog='Play a sine wave over the default output device'
    )
    parser.add_argument('--freq', default=442.0, help='Note frequency (optional)')
    parser.add_argument('--backend', help='Backend to use (optional)')
    parser.add_argument('--blocksize', type=int, help='Block size (optional)')
    parser.add_argument('--rate', type=int, default=44100, help='Sample rate (optional)')
    parser.add_argument('--device', type=int, help='Output device id (optional)')
    args = parser.parse_args()

    player = Player(args.freq, args.backend, args.device, args.rate, args.blocksize)
    print('Playing...')
    print('CTRL-C to exit')

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print('Exiting...')

    player.close()



"""
PySoundIo

A robust, cross-platform solution for real-time audio.
"""
import re
from setuptools import setup, Extension

vstr = open('pysoundio/__init__.py', 'r').read()
regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
version = re.search(regex, vstr, re.M)

soundio = Extension('_soundiox',
                    sources=['pysoundio/_soundiox.c'],
                    include_dirs=['./pysoundio'],
                    libraries=['soundio'])

setup(
    name='pysoundio',
    version=version.group(1),
    description='Python wrapper for libsoundio',
    long_description='A robust, cross-platform solution for real-time audio',
    license='MIT',
    author='Joe Todd',
    author_email='joextodd@gmail.com',
    url='http://pysoundio.readthedocs.io/en/latest/',
    download_url='https://github.com/joextodd/pysoundio/archive/' + version.group(1) + '.tar.gz',
    packages=['pysoundio'],
    ext_modules=[soundio],
    test_suite='tests',
    keywords=('audio', 'sound', 'stream'),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License'
    ],
)



"""
pysoundio.py

Play and Record Sound in Python using libsoundio

libsoundio is a C library for cross-platform audio input and output.
It is suitable for real-time and consumer software.

"""
import logging
import threading

import ctypes as _ctypes
from .constants import (
    DEFAULT_RING_BUFFER_DURATION,
    PRIORITISED_FORMATS,
    PRIORITISED_SAMPLE_RATES,
    SoundIoFormat
)
from .structures import (
    SoundIoErrorCallback,
    SoundIoOverflowCallback,
    SoundIoReadCallback,
    SoundIoUnderflowCallback,
    SoundIoWriteCallback,
    SoundIoDevice,
    SoundIoInStream,
    SoundIoOutStream,
    SoundIoChannelLayout,
    SoundIoSampleRateRange
)
import _soundiox as soundio

LOGGER = logging.getLogger(__name__)


class PySoundIoError(Exception):
    pass


class _InputProcessingThread(threading.Thread):

    def __init__(self, parent, *args, **kwargs):
        self.buffer = parent.input['buffer']
        self.callback = parent.input['read_callback']
        self.bytes_per_frame = parent.input['bytes_per_frame']
        super(_InputProcessingThread, self).__init__(*args, **kwargs)

    def run(self):
        """ Callback with data """
        fill_bytes = soundio.ring_buffer_fill_count(self.buffer)
        read_buf = soundio.ring_buffer_read_ptr(self.buffer)
        if self.callback and fill_bytes:
            self.callback(data=read_buf, length=fill_bytes / self.bytes_per_frame)
        soundio.ring_buffer_advance_read_ptr(self.buffer, fill_bytes)


class _OutputProcessingThread(threading.Thread):

    def __init__(self, parent, block_size, *args, **kwargs):
        self.buffer = parent.output['buffer']
        self.callback = parent.output['write_callback']
        self.bytes_per_frame = parent.output['bytes_per_frame']
        self.block_size = block_size
        super(_OutputProcessingThread, self).__init__(*args, **kwargs)

    def run(self):
        """ Callback to fill data """
        data = bytearray(b'\x00' * self.block_size * self.bytes_per_frame)
        free_bytes = soundio.ring_buffer_free_count(self.buffer)
        if self.callback and free_bytes >= len(data):
            self.callback(data=data, length=self.block_size)
        soundio.ring_buffer_write_ptr(self.buffer, data, len(data))
        soundio.ring_buffer_advance_write_ptr(self.buffer, len(data))


class PySoundIo(object):

    def __init__(self, backend=None):
        """
        Initialise PySoundIo.
        Connect to a specific backend, or the default.

        Parameters
        ----------
        backend: (SoundIoBackend) see `Backends`_. (optional)
        """
        self.backend = backend
        self.testing = False

        self.input = {'device': None, 'stream': None, 'buffer': None, 'read_callback': None}
        self.output = {'device': None, 'stream': None, 'buffer': None, 'write_callback': None}

        self._soundio = soundio.create()
        if backend:
            soundio.connect_backend(backend)
        else:
            soundio.connect()
        self.flush()

    def close(self):
        """
        Clean up allocated memory
        Close libsoundio connections
        """
        if self.input['stream']:
            soundio.instream_destroy()
            self.input['stream'] = None
        if self.output['stream']:
            soundio.outstream_destroy()
            self.output['stream'] = None
        if self.input['buffer']:
            soundio.ring_buffer_destroy(self.input['buffer'])
            self.input['buffer'] = None
        if self.output['buffer']:
            soundio.ring_buffer_destroy(self.output['buffer'])
            self.output['buffer'] = None
        if self.input['device']:
            soundio.device_unref(self.input['device'])
            self.input['device'] = None
        if self.output['device']:
            soundio.device_unref(self.output['device'])
            self.output['device'] = None
        if self._soundio:
            soundio.disconnect()
            soundio.destroy()
            self._soundio = None

    def flush(self):
        """
        Atomically update information for all connected devices.
        """
        soundio.flush()

    @property
    def backend_count(self):
        """
        Returns the number of available backends.
        """
        return soundio.backend_count()

    def get_default_input_device(self):
        """
        Returns default input device

        Returns
        -------
        PySoundIoDevice input device

        Raises
        ------
        PySoundIoError if the input device is not available
        """
        device_id = soundio.default_input_device_index()
        return self.get_input_device(device_id)

    def get_input_device(self, device_id):
        """
        Return an input device by index

        Parameters
        ----------
        device_id: (int) input device index

        Returns
        -------
        PySoundIoDevice input device

        Raises
        ------
        PySoundIoError if an invalid device id is used, or device is unavailable
        """
        if device_id < 0 or device_id > soundio.get_input_device_count():
            raise PySoundIoError('Invalid input device id')
        self.input['device'] = soundio.get_input_device(device_id)
        return self.input['device']

    def get_default_output_device(self):
        """
        Returns default output device

        Returns
        -------
        PySoundIoDevice output device

        Raises
        ------
        PySoundIoError if the output device is not available
        """
        device_id = soundio.default_output_device_index()
        return self.get_output_device(device_id)

    def get_output_device(self, device_id):
        """
        Return an output device by index

        Parameters
        ----------
        device_id: (int) output device index

        Returns
        -------
        PySoundIoDevice output device

        Raises
        ------
        PySoundIoError if an invalid device id is used, or device is unavailable
        """
        if device_id < 0 or device_id > soundio.get_output_device_count():
            raise PySoundIoError('Invalid output device id')
        self.output['device'] = soundio.get_output_device(device_id)
        return self.output['device']

    def list_devices(self):
        """
        Return a list of available devices

        Returns
        -------
        (list)(dict) containing information on available input / output devices.
        """
        output_count = soundio.get_output_device_count()
        input_count = soundio.get_input_device_count()

        default_output = soundio.default_output_device_index()
        default_input = soundio.default_input_device_index()

        input_devices = []
        output_devices = []

        for i in range(0, input_count):
            device = soundio.get_input_device(i)
            pydevice = _ctypes.cast(device, _ctypes.POINTER(SoundIoDevice))
            input_devices.append({
                'id': pydevice.contents.id.decode(), 'name': pydevice.contents.name.decode(),
                'is_raw': pydevice.contents.is_raw, 'is_default': default_input == i,
                'sample_rates': self.get_sample_rates(device),
                'formats': self.get_formats(device),
                'layouts': self.get_layouts(device),
                'software_latency_min': pydevice.contents.software_latency_min,
                'software_latency_max': pydevice.contents.software_latency_max,
                'software_latency_current': pydevice.contents.software_latency_current
            })
            soundio.device_unref(device)

        for i in range(0, output_count):
            device = soundio.get_output_device(i)
            pydevice = _ctypes.cast(device, _ctypes.POINTER(SoundIoDevice))
            output_devices.append({
                'id': pydevice.contents.id.decode(), 'name': pydevice.contents.name.decode(),
                'is_raw': pydevice.contents.is_raw, 'is_default': default_output == i,
                'sample_rates': self.get_sample_rates(device),
                'formats': self.get_formats(device),
                'layouts': self.get_layouts(device),
                'software_latency_min': pydevice.contents.software_latency_min,
                'software_latency_max': pydevice.contents.software_latency_max,
                'software_latency_current': pydevice.contents.software_latency_current
            })
            soundio.device_unref(device)

        LOGGER.info('%d devices found' % (input_count + output_count))
        return (input_devices, output_devices)

    def get_layouts(self, device):
        """
        Return a list of available layouts for a device

        Parameters
        ----------
        device: (SoundIoDevice) device object

        Returns
        -------
        (dict) Dictionary of available channel layouts for a device
        """
        pydevice = _ctypes.cast(device, _ctypes.POINTER(SoundIoDevice))
        current = pydevice.contents.current_layout
        layouts = {
            'current': {'name': current.name.decode() if current.name else 'None'},
            'available': []
        }
        for l in range(0, pydevice.contents.layout_count):
            layouts['available'].append({
                'name': (pydevice.contents.layouts[l].name.decode() if
                    pydevice.contents.layouts[l].name else 'None'),
                'channel_count': pydevice.contents.layouts[l].channel_count
            })
        return layouts

    def get_sample_rates(self, device):
        """
        Return a list of available sample rates for a device

        Parameters
        ----------
        device: (SoundIoDevice) device object

        Returns
        -------
        (dict) Dictionary of available sample rates for a device
        """
        pydevice = _ctypes.cast(device, _ctypes.POINTER(SoundIoDevice))
        sample_rates = {'current': pydevice.contents.sample_rate_current, 'available': []}
        for s in range(0, pydevice.contents.sample_rate_count):
            sample_rates['available'].append({
                'min': pydevice.contents.sample_rates[s].min,
                'max': pydevice.contents.sample_rates[s].max
            })
        return sample_rates

    def get_formats(self, device):
        """
        Return a list of available formats for a device

        Parameters
        ----------
        device: (SoundIoDevice) device object

        Returns
        -------
        (dict) Dictionary of available formats for a device
        """
        pydevice = _ctypes.cast(device, _ctypes.POINTER(SoundIoDevice))
        formats = {'current': pydevice.contents.current_format, 'available': []}
        for r in range(0, pydevice.contents.format_count):
            formats['available'].append(SoundIoFormat[pydevice.contents.formats[r]])
        return formats

    def supports_sample_rate(self, device, rate):
        """
        Check the sample rate is supported by the selected device.

        Parameters
        ----------
        device: (SoundIoDevice) device object
        rate (int): sample rate

        Returns
        -------
        (bool) True if sample rate is supported for this device
        """
        return soundio.device_supports_sample_rate(device, rate) > 0

    def get_default_sample_rate(self, device):
        """
        Get the best sample rate.

        Parameters
        ----------
        device: (SoundIoDevice) device object

        Returns
        -------
        (int) The best available sample rate
        """
        sample_rate = None
        for rate in PRIORITISED_SAMPLE_RATES:
            if self.supports_sample_rate(device, rate):
                sample_rate = rate
                break
        if not sample_rate:
            pydevice = _ctypes.cast(device, _ctypes.POINTER(SoundIoDevice))
            sample_rate = pydevice.contents.sample_rates.contents.max
        return sample_rate

    def supports_format(self, device, format):
        """
        Check the format is supported by the selected device.

        Parameters
        ----------
        device: (SoundIoDevice) device object
        format: (SoundIoFormat) see `Formats`_.

        Returns
        -------
        (bool) True if the format is supported for this device
        """
        return soundio.device_supports_format(device, format) > 0

    def get_default_format(self, device):
        """
        Get the best format value.

        Parameters
        ----------
        device: (SoundIoDevice) device object

        Returns
        ------
        (SoundIoFormat) The best available format
        """
        dtype = soundio.SoundIoFormatInvalid
        for fmt in PRIORITISED_FORMATS:
            if self.supports_format(device, fmt):
                dtype = fmt
                break
        if dtype == soundio.SoundIoFormatInvalid:
            raise PySoundIoError('Incompatible sample formats')
        return dtype

    def sort_channel_layouts(self, device):
        """
        Sorts channel layouts by channel count, descending

        Parameters
        ----------
        device: (SoundIoDevice) device object
        """
        soundio.device_sort_channel_layouts(device)

    def _get_default_layout(self, channels):
        """
        Get default builtin channel layout for the given number of channels

        Parameters
        ----------
        channels: (int) desired number of channels
        """
        return soundio.channel_layout_get_default(channels)

    def get_bytes_per_frame(self, device, channels):
        """
        Get the number of bytes per frame

        Parameters
        ----------
        device: (PySoundIoDevice) device
        channels: (int) number of channels

        Returns
        -------
        (int) number of bytes per frame
        """
        return soundio.get_bytes_per_frame(device, channels)

    def get_bytes_per_sample(self, device):
        """
        Get the number of bytes per sample

        Parameters
        ----------
        device: (PySoundIoDevice) device

        Returns
        -------
        (int) number of bytes per sample
        """
        return soundio.get_bytes_per_sample(device)

    def get_bytes_per_second(self, device, channels, sample_rate):
        """
        Get the number of bytes per second

        Parameters
        ----------
        device: (PySoundIoDevice) device
        channels (int) number of channels
        sample_rate (int) sample rate

        Returns
        -------
        (int) number of bytes per second
        """
        return soundio.get_bytes_per_second(device, channels, sample_rate)

    def _create_input_ring_buffer(self, capacity):
        """
        Creates ring buffer with the capacity to hold 30 seconds of data,
        by default.
        """
        self.input['buffer'] = soundio.input_ring_buffer_create(capacity)
        return self.input['buffer']

    def _create_output_ring_buffer(self, capacity):
        """
        Creates ring buffer with the capacity to hold 30 seconds of data,
        by default.
        """
        self.output['buffer'] = soundio.output_ring_buffer_create(capacity)
        return self.output['buffer']

    def _create_input_stream(self):
        """
        Allocates memory and sets defaults for input stream
        """
        self.input['stream'] = soundio.instream_create(self.input['device'])

        pyinstream = _ctypes.cast(self.input['stream'], _ctypes.POINTER(SoundIoInStream))
        soundio.set_read_callbacks(self._read_callback, self._overflow_callback)

        layout = self._get_default_layout(self.input['channels'])
        pylayout = _ctypes.cast(layout, _ctypes.POINTER(SoundIoChannelLayout))
        pyinstream.contents.layout = pylayout.contents

        pyinstream.contents.format = self.input['format']
        pyinstream.contents.sample_rate = self.input['sample_rate']
        if self.input['block_size']:
            pyinstream.contents.software_latency = float(self.input['block_size']) / self.input['sample_rate']

        return self.input['stream']

    def _open_input_stream(self):
        """
        Open an input stream.
        """
        soundio.instream_open()

    def _start_input_stream(self):
        """
        Start an input stream running.
        """
        soundio.instream_start()

    def pause_input_stream(self, pause):
        """
        Pause input stream

        Parameters
        ----------
        pause: (bool) True to pause, False to unpause
        """
        soundio.instream_pause(pause)

    def get_input_latency(self, out_latency):
        """
        Obtain the number of seconds that the next frame of sound
        being captured will take to arrive in the buffer,
        plus the amount of time that is represented in the buffer.

        Parameters
        ----------
        out_latency: (int) output latency in seconds
        """
        return soundio.instream_get_latency(out_latency)

    def _read_callback(self):
        """
        Internal read callback.
        """
        _InputProcessingThread(parent=self).start()

    def _overflow_callback(self):
        """
        Internal overflow callback, which calls the external
        overflow callback if defined.
        """
        if self.input['overflow_callback']:
            self.input['overflow_callback']()

    def start_input_stream(self, device_id=None,
                           sample_rate=None, dtype=None,
                           block_size=None, channels=None,
                           read_callback=None, overflow_callback=None):
        """
        Creates input stream, and sets parameters. Then allocates
        a ring buffer and starts the stream.

        The read callback is called in an audio processing thread,
        when a block of data is read from the microphone. Data is
        passed from the ring buffer to the callback to process.

        Parameters
        ----------
        device_id: (int) input device id
        sample_rate: (int) desired sample rate (optional)
        dtype: (SoundIoFormat) desired format, see `Formats`_. (optional)
        block_size: (int) desired block size (optional)
        channels: (int) number of channels [1: mono, 2: stereo] (optional)
        read_callback: (fn) function to call with data, the function must have
                        the arguments data and length. See record example
        overflow_callback: (fn) function to call if data is not being read fast enough

        Raises
        ------
        PySoundIoError if any invalid parameters are used

        Notes
        -----
        An example read callback

        .. code-block:: python
            :linenos:

            def read_callback(data, length):
                wav.write(data)

        Overflow callback example

        .. code-block:: python
            :linenos:

            def overflow_callback():
                print('buffer overflow')
        """
        self.input['sample_rate'] = sample_rate
        self.input['format'] = dtype
        self.input['block_size'] = block_size
        self.input['channels'] = channels
        self.input['read_callback'] = read_callback
        self.input['overflow_callback'] = overflow_callback

        if device_id is not None:
            self.input['device'] = self.get_input_device(device_id)
        else:
            self.input['device'] = self.get_default_input_device()

        pydevice = _ctypes.cast(self.input['device'], _ctypes.POINTER(SoundIoDevice))
        LOGGER.info('Input Device: %s' % pydevice.contents.name.decode())
        self.sort_channel_layouts(self.input['device'])

        if self.input['sample_rate']:
            if not self.supports_sample_rate(self.input['device'], self.input['sample_rate']):
                raise PySoundIoError('Invalid sample rate: %d' % self.input['sample_rate'])
        else:
            self.input['sample_rate'] = self.get_default_sample_rate(self.input['device'])

        if self.input['format']:
            if not self.supports_format(self.input['device'], self.input['format']):
                raise PySoundIoError('Invalid format: %s interleaved' %
                                     (soundio.format_string(self.input['format'])))
        else:
            self.input['format'] = self.get_default_format(self.input['device'])

        self._create_input_stream()
        self._open_input_stream()
        pystream = _ctypes.cast(self.input['stream'], _ctypes.POINTER(SoundIoInStream))
        self.input['bytes_per_frame'] = pystream.contents.bytes_per_frame
        capacity = (DEFAULT_RING_BUFFER_DURATION *
                    pystream.contents.sample_rate * pystream.contents.bytes_per_frame)
        self._create_input_ring_buffer(capacity)
        self._start_input_stream()
        self.flush()

    def _create_output_stream(self):
        """
        Allocates memory and sets defaults for output stream
        """
        self.output['stream'] = soundio.outstream_create(self.output['device'])

        pystream = _ctypes.cast(self.output['stream'], _ctypes.POINTER(SoundIoOutStream))
        if not self.testing:
            soundio.set_write_callbacks(self._write_callback, self._underflow_callback)

        layout = self._get_default_layout(self.output['channels'])
        pylayout = _ctypes.cast(layout, _ctypes.POINTER(SoundIoChannelLayout))
        pystream.contents.layout = pylayout.contents

        pystream.contents.format = self.output['format']
        pystream.contents.sample_rate = self.output['sample_rate']
        if self.output['block_size']:
            pystream.contents.software_latency = float(self.output['block_size']) / self.output['sample_rate']

        return self.output['stream']

    def _open_output_stream(self):
        """
        Open an output stream.
        """
        soundio.outstream_open()
        pystream = _ctypes.cast(self.output['stream'], _ctypes.POINTER(SoundIoOutStream))
        self.output['block_size'] = int(pystream.contents.software_latency / self.output['sample_rate'])

    def _start_output_stream(self):
        """
        Start an output stream running.
        """
        soundio.outstream_start()

    def pause_output_stream(self, pause):
        """
        Pause output stream

        Parameters
        ----------
        pause: (bool) True to pause, False to unpause
        """
        soundio.outstream_pause(pause)

    def _write_callback(self, size):
        """
        Internal write callback.
        """
        _OutputProcessingThread(parent=self, block_size=size).start()

    def _underflow_callback(self):
        """
        Internal underflow callback, which calls the external
        underflow callback if defined.
        """
        if self.output['underflow_callback']:
            self.output['underflow_callback']()

    def _clear_output_buffer(self):
        """
        Clear the output buffer
        """
        if self.output['buffer']:
            soundio.ring_buffer_clear(self.output['buffer'])

    def get_output_latency(self, out_latency):
        """
        Obtain the total number of seconds that the next frame written
        will take to become audible.

        Parameters
        ----------
        out_latency: (int) output latency in seconds
        """
        return soundio.outstream_get_latency(out_latency)

    def start_output_stream(self, device_id=None,
                            sample_rate=None, dtype=None,
                            block_size=None, channels=None,
                            write_callback=None, underflow_callback=None):
        """
        Creates output stream, and sets parameters. Then allocates
        a ring buffer and starts the stream.

        The write callback is called in an audio processing thread,
        when a block of data should be passed to the speakers. Data is
        added to the ring buffer to process.

        Parameters
        ----------
        device_id: (int) output device id
        sample_rate: (int) desired sample rate (optional)
        dtype: (SoundIoFormat) desired format, see `Formats`_. (optional)
        block_size: (int) desired block size (optional)
        channels: (int) number of channels [1: mono, 2: stereo] (optional)
        write_callback: (fn) function to call with data, the function must have
                        the arguments data and length.
        underflow_callback: (fn) function to call if data is not being written fast enough

        Raises
        ------
        PySoundIoError if any invalid parameters are used

        Notes
        -----
        An example write callback

        .. code-block:: python
            :linenos:

            def write_callback(data, length):
                outdata = ar.array('f', [0] * length)
                for value in outdata:
                    outdata = 1.0
                data[:] = outdata.tostring()

        Underflow callback example

        .. code-block:: python
            :linenos:

            def underflow_callback():
                print('buffer underflow')
        """
        self.output['sample_rate'] = sample_rate
        self.output['format'] = dtype
        self.output['block_size'] = block_size
        self.output['channels'] = channels
        self.output['write_callback'] = write_callback
        self.output['underflow_callback'] = underflow_callback

        if device_id is not None:
            self.output['device'] = self.get_output_device(device_id)
        else:
            self.output['device'] = self.get_default_output_device()

        pydevice = _ctypes.cast(self.output['device'], _ctypes.POINTER(SoundIoDevice))
        LOGGER.info('Input Device: %s' % pydevice.contents.name.decode())
        self.sort_channel_layouts(self.output['device'])

        if self.output['sample_rate']:
            if not self.supports_sample_rate(self.output['device'], self.output['sample_rate']):
                raise PySoundIoError('Invalid sample rate: %d' % self.output['sample_rate'])
        else:
            self.output['sample_rate'] = self.get_default_sample_rate(self.output['device'])

        if self.output['format']:
            if not self.supports_format(self.output['device'], self.output['format']):
                raise PySoundIoError('Invalid format: %s interleaved' %
                                     (soundio.format_string(self.output['format'])))
        else:
            self.output['format'] = self.get_default_format(self.output['device'])

        self._create_output_stream()
        self._open_output_stream()
        pystream = _ctypes.cast(self.output['stream'], _ctypes.POINTER(SoundIoOutStream))
        self.output['bytes_per_frame'] = pystream.contents.bytes_per_frame
        capacity = (DEFAULT_RING_BUFFER_DURATION *
                    pystream.contents.sample_rate * pystream.contents.bytes_per_frame)
        self._create_output_ring_buffer(capacity)
        self._clear_output_buffer()
        self._start_output_stream()
        self.flush()
